#define NAVIT_KEY_MENU 1 	/* ^A */
#define NAVIT_KEY_LEFT 2	/* ^B */
#define NAVIT_KEY_RIGHT 6	/* ^R */
#define NAVIT_KEY_BACKSPACE 8	/* ^H */
#define NAVIT_KEY_RETURN 13	/* ^M */
#define NAVIT_KEY_DOWN 14	/* ^N */
#define NAVIT_KEY_ZOOM_OUT 15	/* ^O */
#define NAVIT_KEY_UP 16		/* ^P */
#define NAVIT_KEY_ZOOM_IN 17	/* ^Q */
#define NAVIT_KEY_SEARCH 19	/* ^S */
#define NAVIT_KEY_BACK 27	/* Esc */
#define NAVIT_KEY_DELETE 127
#define NAVIT_KEY_TAB 9 /* TAB */
#define NAVIT_KEY_PAGE_UP 33
#define NAVIT_KEY_PAGE_DOWN 34
