# -*- coding: utf-8 -*-
from datetime import datetime

extensions = [
    'm2r',
    'sphinx_rtd_theme',
]

source_suffix = '.rst'

master_doc = 'index'

# General information about the project.
project = u'navit'
year = datetime.now().year
author = u'The Navit Team'
