/**
 * Navit, a modular navigation system.
 * Copyright (C) 2005-2008 Navit Team
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 */

/* This file is generated from http://wiki.navit-project.org/index.php/Item_def.h, do not edit it, edit the wiki page instead */
#ifndef ITEM
#define ITEM(x) extern ##x;
#endif
#ifndef ITEM2
#define ITEM2(x,y) extern ##y;
#endif

ITEM2(0x00000000,none)
ITEM2(0x00000001,point_unspecified)
ITEM(town_streets)
ITEM(street_name)
ITEM(street_name_numbers)
ITEM(street_number)
ITEM(position_sat)
/* Point */
ITEM2(0x00010000,town_label)
ITEM2(0x00010001,town_label_0e0)
ITEM2(0x00010002,town_label_1e0)
ITEM2(0x00010003,town_label_2e0)
ITEM2(0x00010004,town_label_5e0)
ITEM2(0x00010005,town_label_1e1)
ITEM2(0x00010006,town_label_2e1)
ITEM2(0x00010007,town_label_5e1)
ITEM2(0x00010008,town_label_1e2)
ITEM2(0x00010009,town_label_2e2)
ITEM2(0x0001000a,town_label_5e2)
ITEM2(0x0001000b,town_label_1e3)
ITEM2(0x0001000c,town_label_2e3)
ITEM2(0x0001000d,town_label_5e3)
ITEM2(0x0001000e,town_label_1e4)
ITEM2(0x0001000f,town_label_2e4)
ITEM2(0x00010010,town_label_5e4)
ITEM2(0x00010011,town_label_1e5)
ITEM2(0x00010012,town_label_2e5)
ITEM2(0x00010013,town_label_5e5)
ITEM2(0x00010014,town_label_1e6)
ITEM2(0x00010015,town_label_2e6)
ITEM2(0x00010016,town_label_5e6)
ITEM2(0x00010017,town_label_1e7)
ITEM2(0x00010100,district_label)
ITEM2(0x00010101,district_label_0e0)
ITEM2(0x00010102,district_label_1e0)
ITEM2(0x00010103,district_label_2e0)
ITEM2(0x00010104,district_label_5e0)
ITEM2(0x00010105,district_label_1e1)
ITEM2(0x00010106,district_label_2e1)
ITEM2(0x00010107,district_label_5e1)
ITEM2(0x00010108,district_label_1e2)
ITEM2(0x00010109,district_label_2e2)
ITEM2(0x0001010a,district_label_5e2)
ITEM2(0x0001010b,district_label_1e3)
ITEM2(0x0001010c,district_label_2e3)
ITEM2(0x0001010d,district_label_5e3)
ITEM2(0x0001010e,district_label_1e4)
ITEM2(0x0001010f,district_label_2e4)
ITEM2(0x00010110,district_label_5e4)
ITEM2(0x00010111,district_label_1e5)
ITEM2(0x00010112,district_label_2e5)
ITEM2(0x00010113,district_label_5e5)
ITEM2(0x00010114,district_label_1e6)
ITEM2(0x00010115,district_label_2e6)
ITEM2(0x00010116,district_label_5e6)
ITEM2(0x00010117,district_label_1e7)
ITEM(country_label)
ITEM(town_ghost)
ITEM(highway_exit_label)
ITEM(port_label)
ITEM(label_unkn)
ITEM(highway_exit)
ITEM(poi_lake)
ITEM(poi_island)
ITEM(poi)
ITEM(waypoint)
ITEM(trackpoint)
ITEM(bookmark)
ITEM(former_destination)
ITEM(poi_land_feature)
ITEM(poi_cape)
ITEM(poi_rock)
ITEM(poi_airport)
ITEM(poi_toll_booth)
ITEM(poi_fuel)
ITEM(poi_hotel)
ITEM(poi_camp_rv)
ITEM(poi_marina)
ITEM(poi_attraction)
ITEM(poi_museum_history)
ITEM(poi_shopping)
ITEM(poi_car_dealer_parts)
ITEM(poi_car_parking)
ITEM(poi_wreck)
ITEM(poi_building)
ITEM(poi_bridge)
ITEM(poi_park)
ITEM(poi_water_feature)
ITEM(poi_bar)
ITEM(poi_picnic)
ITEM(poi_hospital)
ITEM(poi_camping)
ITEM(poi_public_utilities)
ITEM(poi_burgerking)
ITEM(poi_kfc)
ITEM(poi_mcdonalds)
ITEM(poi_wienerwald)
ITEM(poi_dining)
ITEM(poi_fastfood)
ITEM(poi_police)
ITEM(poi_auto_club)
ITEM(poi_autoservice)
ITEM(poi_bank)
ITEM(poi_bay)
ITEM(poi_bend)
ITEM(poi_boat_ramp)
ITEM(poi_border_station)
ITEM(poi_bowling)
ITEM(poi_bus_station)
ITEM(poi_bus_stop)
ITEM(poi_business_service)
ITEM(poi_car_rent)
ITEM(poi_car_wash)
ITEM(poi_casino)
ITEM(poi_cemetery)
ITEM(poi_church)
ITEM(poi_cinema)
ITEM(poi_civil_removeme)
ITEM(poi_communication)
ITEM(poi_concert)
ITEM(poi_cove)
ITEM(poi_crossing)
ITEM(poi_dam)
ITEM(poi_danger_area)
ITEM(poi_danger_sea_wreck)
ITEM(poi_daymark)
ITEM(poi_diving)
ITEM(poi_drinking_water)
ITEM(poi_emergency)
ITEM(poi_fair)
ITEM(poi_firebrigade)
ITEM(poi_fish)
ITEM(poi_forbidden_area)
ITEM(poi_shop_gps)
ITEM(poi_golf)
ITEM(poi_government_building)
ITEM(poi_height)
ITEM(poi_heliport)
ITEM(poi_hotspring)
ITEM(poi_icesport)
ITEM(poi_information)
ITEM(poi_justice)
ITEM(poi_landmark)
ITEM(poi_levee)
ITEM(poi_library)
ITEM(poi_locale)
ITEM(poi_loudspeaker)
ITEM(poi_mall)
ITEM(poi_manmade_feature)
ITEM(poi_marine)
ITEM(poi_marine_type)
ITEM(poi_mark)
ITEM(poi_military)
ITEM(poi_mine)
ITEM(poi_nondangerous)
ITEM(poi_oil_field)
ITEM(poi_personal_service)
ITEM(poi_pharmacy)
ITEM(poi_post_removeme)
ITEM(poi_public_office)
ITEM(poi_repair_service)
ITEM(poi_resort)
ITEM(poi_rest_room_removeme)
ITEM(poi_restaurant)
ITEM(poi_restricted_area)
ITEM(poi_restroom)
ITEM(poi_sailing)
ITEM(poi_scenic_area)
ITEM(poi_school)
ITEM(poi_service)
ITEM(poi_shop_apparel)
ITEM(poi_shop_computer)
ITEM(poi_shop_department)
ITEM(poi_shop_furnish_removeme)
ITEM(poi_shop_grocery)
ITEM(poi_shop_handg)
ITEM(poi_shop_merchandise)
ITEM(poi_shop_retail)
ITEM(poi_shower)
ITEM(poi_skiing)
ITEM(poi_social_service)
ITEM(poi_sounding)
ITEM(poi_sport)
ITEM(poi_stadium)
ITEM(poi_subdivision_removeme)
ITEM(poi_swimming)
ITEM(poi_telephone)
ITEM(poi_theater)
ITEM(poi_tide)
ITEM(poi_tower)
ITEM(poi_trail)
ITEM(poi_truck_stop)
ITEM(poi_tunnel)
ITEM(poi_wine)
ITEM(poi_worship)
ITEM(poi_wrecker)
ITEM(poi_zoo)
ITEM(rg_point)
ITEM(point_unkn)
ITEM(traffic_signals)
ITEM(poi_gc_multi)
ITEM(poi_gc_tradi)
ITEM(poi_gc_event)
ITEM(poi_gc_mystery)
ITEM(poi_gc_question)
ITEM(poi_gc_stages)
ITEM(poi_gc_reference)
ITEM(poi_gc_webcam)
ITEM(poi_cafe)
ITEM(nav_straight)
ITEM(nav_turnaround)
ITEM(nav_right_1)
ITEM(nav_right_2)
ITEM(nav_right_3)
ITEM(nav_left_1)
ITEM(nav_left_2)
ITEM(nav_left_3)
ITEM(nav_roundabout_r1)
ITEM(nav_roundabout_r2)
ITEM(nav_roundabout_r3)
ITEM(nav_roundabout_r4)
ITEM(nav_roundabout_r5)
ITEM(nav_roundabout_r6)
ITEM(nav_roundabout_r7)
ITEM(nav_roundabout_r8)
ITEM(nav_roundabout_l1)
ITEM(nav_roundabout_l2)
ITEM(nav_roundabout_l3)
ITEM(nav_roundabout_l4)
ITEM(nav_roundabout_l5)
ITEM(nav_roundabout_l6)
ITEM(nav_roundabout_l7)
ITEM(nav_roundabout_l8)
ITEM(poi_peak)
ITEM(poi_rail_station)
ITEM(poi_image)
ITEM(mini_roundabout)
ITEM(turning_circle)
ITEM(poi_townhall)
ITEM(poi_level_crossing)
ITEM(poi_rail_halt)
ITEM(poi_rail_tram_stop)
ITEM(poi_wifi)
ITEM(poi_bench)
ITEM(poi_biergarten)
ITEM(poi_boundary_stone)
ITEM(poi_castle)
ITEM(poi_hunting_stand)
ITEM(poi_memorial)
ITEM(poi_monument)
ITEM(poi_shelter)
ITEM(poi_fountain)
ITEM(poi_potable_water)
ITEM(poi_toilets)
ITEM(poi_viewpoint)
ITEM(poi_ruins)
ITEM(nav_none)
ITEM(nav_position)
ITEM(nav_destination)
ITEM(tec_common)
ITEM(trackpoint_tracked)
ITEM(announcement)
ITEM(poi_post_box)
ITEM(poi_post_office)
ITEM(poi_school_university)
ITEM(poi_school_college)
ITEM(poi_motel)
ITEM(poi_guesthouse)
ITEM(poi_hostel)
ITEM(poi_taxi)
ITEM(poi_prison)
ITEM(poi_kindergarten)
ITEM(poi_shop_butcher)
ITEM(poi_shop_baker)
ITEM(poi_shop_kiosk)
ITEM(poi_soccer)
ITEM(poi_basketball)
ITEM(poi_baseball)
ITEM(poi_climbing)
ITEM(poi_motor_sport)
ITEM(poi_tennis)
ITEM(house_number)
ITEM(route_start)
ITEM(route_end)
ITEM(selected_point)
ITEM(power_tower)
ITEM(route_start_reverse)
ITEM(log_entry)
ITEM(poi_playground)
ITEM(barrier_cycle)
ITEM(barrier_bollard)
ITEM(poi_vending_machine)
ITEM(poi_recycling)
ITEM(poi_hairdresser)
ITEM(poi_shop_fruit)
ITEM(poi_shop_bicycle)
ITEM(poi_shop_florist)
ITEM(poi_shop_optician)
ITEM(poi_shop_beverages)
ITEM(poi_nightclub)
ITEM(poi_shop_shoes)
ITEM(power_substation)
ITEM(barrier_lift_gate)
ITEM(traffic_sign_stop)
ITEM(traffic_sign_city_limit)
ITEM(poi_tree)
ITEM(poi_shop_furniture)
ITEM(poi_shop_parfum)
ITEM(poi_bicycle_rental)
ITEM(poi_bicycle_parking)
ITEM(poi_car_sharing)
ITEM(poi_shop_drugstore)
ITEM(poi_shop_photo)
ITEM(ch_node)
ITEM(place_label)
ITEM(vehicle)
ITEM(vehicle_pedestrian)
ITEM(bookmark_folder)
ITEM(poi_pub)
ITEM(poi_atm)
ITEM(poi_taoist)
ITEM(poi_sikh)
ITEM(poi_shinto)
ITEM(poi_pastafarian)
ITEM(poi_pagan)
ITEM(poi_jewish)
ITEM(poi_jain)
ITEM(poi_islamic)
ITEM(poi_hindu)
ITEM(poi_buddhist)
ITEM(poi_bahai)
ITEM(found_item)
ITEM(nav_merge_left)
ITEM(nav_merge_right)
ITEM(nav_turnaround_left)
ITEM(nav_turnaround_right)
ITEM(nav_exit_left)
ITEM(nav_exit_right)
ITEM(nav_keep_left)
ITEM(nav_keep_right)
ITEM(poi_cave)
ITEM(poi_archaeological_site)
ITEM2(0x7fffffe0,poi_customg)
ITEM(poi_customh)
ITEM(poi_customi)
ITEM(poi_customj)
ITEM(poi_customk)
ITEM(poi_customl)
ITEM(poi_customm)
ITEM(poi_customn)
ITEM(poi_customo)
ITEM(poi_customp)
ITEM(poi_customq)
ITEM(poi_customr)
ITEM(poi_customs)
ITEM(poi_customt)
ITEM(poi_customu)
ITEM(poi_customv)
ITEM(poi_custom0)
ITEM(poi_custom1)
ITEM(poi_custom2)
ITEM(poi_custom3)
ITEM(poi_custom4)
ITEM(poi_custom5)
ITEM(poi_custom6)
ITEM(poi_custom7)
ITEM(poi_custom8)
ITEM(poi_custom9)
ITEM(poi_customa)
ITEM(poi_customb)
ITEM(poi_customc)
ITEM(poi_customd)
ITEM(poi_custome)
ITEM(poi_customf)
/* Line */
ITEM2(0x80000000,line)
ITEM2(0x80000001,line_unspecified)
ITEM(border_country)
ITEM(border_state)
ITEM(rail)
ITEM(water_line)
ITEM(street_nopass)
ITEM(street_0)
ITEM(street_1_city)
ITEM(street_2_city)
ITEM(street_3_city)
ITEM(street_4_city)
ITEM(highway_city)
ITEM(street_1_land)
ITEM(street_2_land)
ITEM(street_3_land)
ITEM(street_4_land)
ITEM(street_n_lanes)
ITEM(highway_land)
ITEM(ramp)
ITEM(roundabout)
ITEM(ferry)
ITEM(roadbook)
ITEM(street_unkn)
ITEM(street_route)
ITEM(height_line_1)
ITEM(height_line_2)
ITEM(track)
ITEM(height_line_3)
ITEM(depth_line_1)
ITEM(depth_line_2)
ITEM(depth_line_3)
ITEM(powerline)
ITEM(pipeline)
ITEM(time_zone)
ITEM(marine_boundary)
ITEM(marine_hazard)
ITEM(rg_segment)
ITEM(aeroway_runway)
ITEM(aeroway_taxiway)
ITEM(street_service)
ITEM(coverage)
ITEM(street_pedestrian)
ITEM(bridge)
ITEM(tunnel)
ITEM(bridleway)
ITEM(cycleway)
ITEM(footway)
ITEM(steps)
ITEM(track_paved)
ITEM(track_gravelled)
ITEM(track_unpaved)
ITEM(track_ground)
ITEM(track_grass)
ITEM(rail_narrow_gauge)
ITEM(rail_light)
ITEM(rail_subway)
ITEM(rail_mono)
ITEM(rail_tram)
ITEM(rail_preserved)
ITEM(rail_disused)
ITEM(rail_abandoned)
ITEM(lift_cable_car)
ITEM(lift_chair)
ITEM(lift_drag)
ITEM(living_street)
ITEM(bus_guideway)
ITEM(street_construction)
ITEM(border_civil)
ITEM(border_political)
ITEM(border_national_park)
ITEM(water_river)
ITEM(water_canal)
ITEM(water_stream)
ITEM(water_drain)
ITEM(path)
ITEM(hiking)
ITEM(hiking_mountain)
ITEM(hiking_mountain_demanding)
ITEM(hiking_alpine)
ITEM(hiking_alpine_demanding)
ITEM(hiking_alpine_difficult)
ITEM(street_parking_lane)
ITEM(piste_nordic)
ITEM(piste_downhill_novice)
ITEM(piste_downhill_easy)
ITEM(piste_downhill_intermediate)
ITEM(piste_downhill_advanced)
ITEM(piste_downhill_expert)
ITEM(piste_downhill_freeride)
ITEM(footway_and_piste_nordic)
ITEM(tracking_0)
ITEM(tracking_10)
ITEM(tracking_20)
ITEM(tracking_30)
ITEM(tracking_40)
ITEM(tracking_50)
ITEM(tracking_60)
ITEM(tracking_70)
ITEM(tracking_80)
ITEM(tracking_90)
ITEM(tracking_100)
ITEM(track_tracked)
ITEM(house_number_interpolation_even)
ITEM(house_number_interpolation_odd)
ITEM(house_number_interpolation_all)
ITEM(house_number_interpolation_alphabetic)
ITEM(ditch)
ITEM(hedge)
ITEM(fence)
ITEM(wall)
ITEM(retaining_wall)
ITEM(city_wall)
ITEM(street_turn_restriction_no)
ITEM(street_turn_restriction_only)
ITEM(border_city)
ITEM(border_county)
ITEM(selected_line)
ITEM(traffic_distortion)
ITEM(street_route_occluded)
ITEM(forest_way_1)
ITEM(forest_way_2)
ITEM(forest_way_3)
ITEM(forest_way_4)
ITEM(former_itinerary)
ITEM(former_itinerary_part)
ITEM(cliff)
ITEM(sports_track)
ITEM(archaeological_site)
/* Area */
ITEM2(0xc0000000,area)
ITEM2(0xc0000001,area_unspecified)
ITEM(poly_wood)
ITEM(poly_water)
ITEM(poly_town)
ITEM(poly_cemetery)
ITEM(poly_car_parking)
ITEM(poly_industry)
ITEM(poly_airport)
ITEM(poly_hospital)
ITEM(poly_park)
ITEM(poly_sport)
ITEM(poly_museum)
ITEM(image)
ITEM(image_path)
ITEM(poly_commercial_center)
ITEM(poly_golf_course)
ITEM(poly_university)
ITEM(poly_national_park)
ITEM(poly_nature_park)
ITEM(poly_flats)
ITEM(poly_scrub)
ITEM(poly_military_zone)
ITEM(poly_marine)
ITEM(plantation)
ITEM(tundra)
ITEM(tile)
ITEM(submap)
ITEM(poly_building)
ITEM(poly_place)
ITEM(poly_station)
ITEM(poly_farm)
ITEM(poly_sport_tennis)
ITEM(poly_apron)
ITEM(poly_terminal)
ITEM(countryindex)
ITEM(poly_sports_centre)
ITEM(poly_sports_stadium)
ITEM(poly_sports_track)
ITEM(poly_sports_pitch)
ITEM(poly_water_park)
ITEM(poly_marina)
ITEM(poly_fishing)
ITEM(poly_theme_park)
ITEM(poly_attraction)
ITEM(poly_ruins)
ITEM(poly_archaeological_site)
ITEM(poly_artwork)
ITEM(poly_zoo)
ITEM(poly_camp_site)
ITEM(poly_caravan_site)
ITEM(poly_picnic_site)
ITEM(poly_playground)
ITEM(poly_allotments)
ITEM(poly_village_green)
ITEM(poly_recreation_ground)
ITEM(poly_common)
ITEM(poly_garden)
ITEM(poly_nature_reserve)
ITEM(poly_glacier)
ITEM(poly_scree)
ITEM(poly_fell)
ITEM(poly_heath)
ITEM(poly_marsh)
ITEM(poly_mud)
ITEM(poly_land)
ITEM(poly_beach)
ITEM(poly_quarry)
ITEM(poly_landfill)
ITEM(poly_retail)
ITEM(poly_commercial)
ITEM(poly_brownfield)
ITEM(poly_greenfield)
ITEM(poly_construction)
ITEM(poly_railway)
ITEM(poly_military)
ITEM(poly_airfield)
ITEM(poly_barracks)
ITEM(poly_danger_area)
ITEM(poly_range)
ITEM(poly_naval_base)
ITEM(poly_basin)
ITEM(poly_reservoir)
ITEM(poly_college)
ITEM(poly_battlefield)
ITEM(poly_pedestrian)
ITEM(poly_plaza)
ITEM(poly_service)
ITEM(poly_street_1)
ITEM(poly_street_2)
ITEM(poly_street_3)
ITEM(poly_wetland)
ITEM(poly_aeroway_runway)
ITEM(map_information)
ITEM(selected_area)
ITEM(poly_place1)
ITEM(poly_place2)
ITEM(poly_place3)
ITEM(poly_place4)
ITEM(poly_place5)
ITEM(poly_place6)
ITEM(poly_water_tiled)
ITEM(poly_meadow)
ITEM(poly_depot)
ITEM(poly_garages)
ITEM(poly_greenhouse)
ITEM(poly_orchard)
ITEM(poly_plantnursery)
ITEM(poly_port)
ITEM(poly_saltpond)
ITEM(poly_dam)
ITEM(poly_swimming_pool)
ITEM2(0xffffffff,last)
